from pysimmods.model.model import Model


class Consumer(Model):
    """A consumer subtype model.

    A consumer returns positive power values in the consumer reference
    arrow system.

    """

    @property
    def set_percent(self):
        p_kw = self.inputs.p_set_kw
        if p_kw is None:
            return None

        p_kw /= self.config.pn_max_kw - self.config.pn_min_kw

        return p_kw * 100

    @set_percent.setter
    def set_percent(self, value):
        value = max(min(abs(value), 100.0), 0.0)
        if value == 0:
            p_set_kw = 0
        else:
            p_range = self.config.pn_max_kw - self.config.pn_min_kw
            p_set_kw = self.config.pn_min_kw + value / 100 * p_range
        self.inputs.p_set_kw = p_set_kw


from pysimmods.model.model import Model


class Buffer(Model):
    """A buffer subtype model.

    This class provides unified access to set_percent for all buffer
    models. With passive sign convention, the buffer returns negative
    power for generation and positive power for consumption. With
    active sign convention, it is the other way around.

    *set_percent* hides the sign convention. In contrast to generator
    and load, the battery is idle with a set_percent of 50.

    """

    @property
    def set_percent(self):
        p_kw = self.inputs.p_set_kw
        if p_kw is None:
            return None

        if p_kw * self.config.gsign > 0:
            # Generating energy
            set_percent = abs(p_kw / self.config.p_discharge_max_kw)
            set_percent *= 50 * self.config.gsign
            return 50 + set_percent
        else:
            # Consuming energy
            set_percent = abs(p_kw / self.config.p_charge_max_kw)
            set_percent *= 50 * self.config.lsign
            return 50 + set_percent

    @set_percent.setter
    def set_percent(self, value):
        value = max(min(abs(value), 100.0), 0.0)

        if value < 50:
            # Generating energy in psc
            p_max_kw = self.config.p_min_kw
        else:
            # Consuming energy in psc
            p_max_kw = self.config.p_max_kw

        set_percent = abs((50 - value) * 2 / 100)
        self.inputs.p_set_kw = set_percent * p_max_kw
        # print()


"""This module contains the input model for the inverter."""


class InverterInputs:
    """Inverter inputs.

    This model does not inherit from the :class:`~.ModelInputs`,
    because those inputs are not required for the inverter.

    Attributes
    ----------
    p_in_kw : float
        Incoming (available) active power in [kW]
    p_set_kw : float
        Target active power in [kW]. Only used in the *'p_set'*,
        *'pq_set'*, and *'qp_set'* modes of the inverter.
    q_set_kw : float
        Target reactive power in [kVAr]. Only used in the *'q_set'*,
        *'pq_set'*, and *'qp_set'* modes of the inverter.
    cos_phi_set : float
        Target cos phi between 0 and 1. Only used in the
        *'cos_phi_set'* mode of the inverter.

    """

    def __init__(self):
        self.p_in_kw = None
        self.p_set_kw = None
        self.q_set_kvar = None
        self.cos_phi_set = None

    def reset(self):
        self.p_in_kw = None
        self.p_set_kw = None
        self.q_set_kvar = None
        self.cos_phi_set = None



from pysimmods.model.model import Model
from pysimmods.model.config import ModelConfig
from pysimmods.model.state import ModelState
from pysimmods.model.inputs import ModelInputs


class DummyConfig(ModelConfig):
    def __init__(self, params):
        super().__init__(params)

        self._default_schedule = [0 for _ in range(24)]


class DummyState(ModelState):
    def __init__(self, inits):
        super().__init__(inits)


class DummyInputs(ModelInputs):
    pass


class DummyModel(Model):
    def __init__(self, params, inits):
        self.config = DummyConfig(params)
        self.state = DummyState(inits)
        self.inputs = DummyInputs()

    def step(self):
        self.state.p_kw = self.inputs.p_set_kw

    @property
    def set_percent(self):
        return self.inputs.p_set_kw

    @set_percent.setter
    def set_percent(self, value):
        self.inputs.p_set_kw = value



"""This module contains the base config for each model."""


class ModelConfig:
    """The base config for all models.

    Parameters
    ----------
    params : dict
        A *dict* containing configuration parameters. See *Attributes*
        for detailed information

    Attributes
    ----------
    sign_convention : str
        Should be defined in *params* and can have the values
        *'passive'* or *'active'*. Passive sign convention aka load
        reference arrow system is normally used and leads to positive
        consumption and negative generation power flows.
    psc : bool
        Is *True* if passive sign convention is used.
    asc : bool
        Is *True* if active sign convention is used.
    gsign : int
        The generator sign depending on the sign convention. Will be
        set to *-1* if passive sign convention is used and to *1*
        otherwise.
    lsign : int
        The load sign depending on the sign convention. Will be set to
        *1* is passive sign convention is used and to *-1* otherwise.

    """

    def __init__(self, params):

        self.sign_convention = params.get("sign_convention", "passive")
        assert self.sign_convention in ["active", "passive"]

        self.psc = self.sign_convention == "passive"
        self.asc = self.sign_convention == "active"

        if self.sign_convention == "active":
            self.gsign = 1
            self.lsign = -1
        else:
            self.gsign = -1
            self.lsign = 1

        self._default_schedule = [50] * 24

    @property
    def p_max_kw(self):
        raise NotImplementedError()

    @property
    def p_min_kw(self):
        raise NotImplementedError()

    @property
    def q_max_kvar(self):
        raise NotImplementedError()

    @property
    def q_min_kvar(self):
        raise NotImplementedError()

    @property
    def default_schedule(self):
        """Return the default schedule.

        Returns
        -------
        list
            A *list* containing a default schedule with percentage
            power for each hour of the day (i.e.,
            len(default_schedule) == 24). This is used if no other *p*
            or *q* input is provided.

        """
        return self._default_schedule



"""This module contains the flexibility model."""
from datetime import datetime, timedelta

import numpy as np

import pandas as pd
from pysimmods.util.dateformat import GER

from .forecast_model import ForecastModel


class FlexibilityModel(ForecastModel):
    """The flexibility model for all pysimmods."""

    def generate_schedules(self, start, forecast_horizon_hours, num_schedules):
        """Perform sampling and generate a set of schedules for the
        specified time interval.

        Args:
            start (str): Is the start of the planning horizon for which
                the sampling is done. It has to be provided as ISO 8601
                timezone string such as '2020-06-22 12:00:00+0000'

            forecast_horizon_hours (int): The planning horizon is divided into.

        """
        state_backup = self.model.get_state()
        step_size = self.step_size

        now_dt = self.now_dt
        start_dt = datetime.strptime(start, GER)
        end_dt = (
            start_dt
            + timedelta(hours=forecast_horizon_hours)
            - timedelta(seconds=step_size)
        )
        periods = forecast_horizon_hours * 3_600 / step_size

        # Fast forward to the planning interval
        while start_dt > now_dt:
            try:
                self._perform_step(
                    now_dt,
                    self.schedule.get("target", now_dt) / self.percent_factor,
                )
                now_dt += timedelta(seconds=step_size)
            except KeyError:
                return dict()

        index = pd.date_range(start_dt, end_dt, periods=periods)

        self.flexibilities = dict()

        for schedule_id in range(num_schedules):
            schedule = self._generate_schedule(
                pd.DataFrame(
                    columns=["target", self.pname, self.qname], index=index
                )
            )

            self.flexibilities[schedule_id] = schedule

        self.model.set_state(state_backup)
        return self.flexibilities

    def _generate_schedule(self, dataframe):
        num_steps = len(dataframe.index)
        dataframe["target"] = self.rng.uniform(size=num_steps) / (
            0.01 / self.percent_factor
        )

        state_backup = self.model.get_state()
        for index, row in dataframe.iterrows():
            try:
                self._perform_step(index, row["target"])
                dataframe.loc[index, self.pname] = (
                    self.model.p_kw * self.unit_factor
                )
                dataframe.loc[index, self.qname] = (
                    self.model.q_kvar * self.unit_factor
                )

            except KeyError:
                # Forecast is missing
                dataframe.loc[index, self.pname] = np.nan
                dataframe.loc[index, self.qname] = np.nan
                dataframe.loc[index, "target"] = np.nan
        self.model.set_state(state_backup)

        return dataframe



from .model import Model


class Generator(Model):
    """A generator subtype model.

    This class provides unified access to set_percent for all
    generators. A generator returns negative power values in the
    consumer reference arrow system (passive sign convention), which
    is hided if one uses *set_percent*, i.e. calling set_percent
    with, e.g., a value of 50 will always set the power of the
    generator to 50 percent, independently of the reference system
    used.

    """

    @property
    def set_percent(self):
        return abs(self.inputs.p_set_kw / (self.p_min_kw)) * 100.0

    @set_percent.setter
    def set_percent(self, set_percent):
        set_percent = max(min(abs(set_percent * 0.01), 1.0), 0.0)
        val_range = self.p_min_kw - self.p_max_kw
        self.inputs.p_set_kw = self.p_max_kw + set_percent * val_range



"""This module contains the config model of the PV plant system."""
import math

from pysimmods.other.invertersim.config import InverterConfig
from pysimmods.model.config import ModelConfig
from pysimmods.generator.pvsim.config import PVConfig


class PVSystemConfig(ModelConfig):
    """Config parameters of the PV plant System.

    Consists of a :class:`~.PVConfig` and an :class:`~.InverterConfig`
    object.

    """

    def __init__(self, params):
        super().__init__(params)
        params["pv"]["sign_convention"] = self.sign_convention
        params["inverter"]["sign_convention"] = self.sign_convention

        self.pv = PVConfig(params["pv"])
        self.inverter = InverterConfig(params["inverter"])

    @property
    def p_min_kw(self):
        """Minimum active power of pv system in kW, this property is
        used by :class:`.mosaik.flexibility_model.FlexibilityModel`.

        With passive sign convention, this is the maximum power this
        plant can generate.

        """
        if self.psc:
            return self._get_abs_p_max()
        else:
            return 0

    @property
    def p_max_kw(self):
        """Maximum active power of pv plant in kW, this property is
        used by :class:`.mosaik.flexibility_model.FlexibilityModel`.

        With active sign convention, this is the maximum power this
        plant can generate

        """
        if self.psc:
            return 0
        else:
            return self._get_abs_p_max()

    @property
    def q_min_kvar(self):
        """Minimum reactive power of pv plant in kVAr.

        For certain combinations, this is the maximum absolute power
        this plant can generate.

        """
        if self.psc and self.inverter.inverter_mode == "capacitive":
            return self._get_abs_q_max()
        elif self.asc and self.inverter.inverter_mode == "inductive":
            return self._get_abs_q_max()
        else:
            return 0

    @property
    def q_max_kvar(self):
        """Maximum reactive power of pv plant in kVAr.

        For certain combinations, this is the maximum absolute power
        this plant can generate.

        """
        if self.psc and self.inverter.inverter_mode == "inductive":
            return self._get_abs_q_max()
        elif self.asc and self.inverter.inverter_mode == "capacitive":
            return self._get_abs_q_max()
        else:
            return 0

    @property
    def default_schedule(self):
        return self.pv.default_schedule

    def _get_abs_p_max(self):
        if self.pv.p_peak_kw > self.inverter.sn_kva:
            return self.inverter.sn_kva * self.gsign
        else:
            return self.pv.p_peak_kw * self.gsign

    def _get_abs_q_max(self):
        const_cos_phi_modes = ["p_set", "q_set"]
        variable_cos_phi_modes = ["cos_phi_set", "pq_set", "qp_set"]

        if self.inverter.q_control in const_cos_phi_modes:
            q_max_kvar = self.inverter.sn_kva * math.sin(
                math.acos(self.inverter.cos_phi)
            )

            return q_max_kvar * self.gsign

        elif self.inverter.q_control in variable_cos_phi_modes:
            return self.inverter.sn_kva * self.gsign

        else:
            return 0



"""This module contains the config model for the battery model."""
from pysimmods.model.config import ModelConfig


class BatteryConfig(ModelConfig):
    """Captures the configuration parameters of the battery model

    On intitialization a dictionay with values for all configuration
    parameters has to be passed.

    The configuration parameters are constant during the simulation
    process. That is they are not manipulated in the step-method of
    the battery model and should not be changed during simulation
    from outside.

    Attributes
    ----------
    cap_kwh : float
        Capacity of the battery in [kWh].
    p_charge_max_kw : float
        Maximum charging (consumption) power of battery in [kW].
    p_discharge_max_kw : float
        Maximum discharging (generation) power of battery in [kW].
    soc_min_percent : float
        Minimum state of charge of battery in [%] of capacity.
    eta_pc : list
        Polynomial coefficients for calculating set power dependent eta.
    """

    def __init__(self, params):
        super().__init__(params)

        self.cap_kwh = abs(params["cap_kwh"])
        self.p_charge_max_kw = abs(params["p_charge_max_kw"])
        self.p_discharge_max_kw = abs(params["p_discharge_max_kw"])
        self.soc_min_percent = params["soc_min_percent"]
        self.eta_pc = params["eta_pc"]

        self._default_schedule = [
            50,
            59.451814,
            60,
            60,
            60,
            60,
            50,
            50,
            50,
            35,
            35,
            35,
            35,
            50,
            50,
            70,
            70,
            50,
            35,
            35,
            50,
            51,
            51,
            54,
        ]

    @property
    def p_min_kw(self):
        """Minimum power of battery in kW, this property is used by
        :meth:`.model.Model.get_sample`

        """
        if self.psc:
            return self.p_discharge_max_kw * self.gsign
        else:
            return self.p_charge_max_kw * self.lsign

    @property
    def p_max_kw(self):
        """Maximum power of battery in kW, this property is used by
        :meth:`.model.Model.get_sample`

        """
        if self.psc:
            return self.p_charge_max_kw * self.lsign
        else:
            return self.p_discharge_max_kw * self.gsign

    @property
    def q_min_kvar(self):
        """Needs to be implemented."""
        return 0

    @property
    def q_max_kvar(self):
        """Needs to be implemented."""
        return 0



"""This module contains a :class:`mosaik_api.Simulator` for the 
flexiblity model, which is a wrapper for all models of the 
pysimmods package.

"""
from datetime import datetime, timedelta

import mosaik_api
import numpy as np
from pysimmods.mosaik import LOG
from pysimmods.util.dateformat import GER
from pysimmods.other.flexibility.flexibility_model import FlexibilityModel
from .meta import META, MODELS


class FlexibilitySimulator(mosaik_api.Simulator):
    """The generic flexiblity mosaik simulator for all pysimmods."""

    def __init__(self):
        super().__init__(META)
        self.sid = None
        self.models = dict()
        self.num_models = dict()

        self.step_size = None
        self.now_dt = None

        self.planning_horizon = None
        self.num_schedules = 0

    def init(self, sid, **sim_params):
        """Called exactly ones after the simulator has been started.

        Parameters
        ----------
        sid : str
            Simulator ID for this simulator.
        start_date : str
            The start date as UTC ISO 8601 date string.
        step_size : int, optional
            Step size for this simulator. Defaults to 900.

        Returns
        -------
        dict
            The meta dict (set by *mosaik_api.Simulator*).

        """
        self.sid = sid
        if "step_size" not in sim_params:
            LOG.debug(
                "Param *step_size* not provided. "
                "Using default step size of 900."
            )
        self.step_size = sim_params.get("step_size", 900)
        self.now_dt = datetime.strptime(sim_params["start_date"], GER)

        self.unit = sim_params.get("unit", "kw")
        self.forecast_horizon_hours = sim_params.get(
            "forecast_horizon_hours", self.step_size / 3_600
        )

        self.provide_flexibilities = sim_params.get(
            "provide_flexibilities", False
        )
        self.flexibility_horizon_hours = sim_params.get(
            "flexibility_horizon_hours", 2
        )
        self.num_schedules = sim_params.get("num_schedules", 10)
        self.rng = np.random.RandomState(sim_params.get("seed", None))

        self._update_meta()

        return self.meta

    def create(self, num, model, **model_params):
        """Initialize the simulation model instance (entity).

        Parameters
        ----------
        num : int
            The number of models to create.
        model : str
            The name of the models to create. Must be present inside
            the simulator's meta.

        Returns
        -------
        list
            A list with information on the created entity.

        """
        entities = []
        params = model_params["params"]
        inits = model_params["inits"]
        self.num_models.setdefault(model, 0)

        for _ in range(num):
            eid = f"{model}-{self.num_models[model]}"
            self.models[eid] = FlexibilityModel(
                MODELS[model](params, inits),
                self.now_dt,
                self.step_size,
                self.forecast_horizon_hours,
                self.flexibility_horizon_hours,
                self.num_schedules,
                self.rng.randint(1_000_000),
                self.unit,
            )
            self.num_models[model] += 1
            entities.append({"eid": eid, "type": model})
        return entities

    def step(self, time, inputs):
        """Perform a simulation step.

        Parameters
        ----------
        time : int
            The current simulation step (by convention in seconds since
            simulation start.
        inputs : dict
            A *dict* containing inputs for entities of this simulator.

        Returns
        -------
        int
            The next step this simulator wants to be stepped.

        """
        LOG.debug("At step %d: Received inputs: %s.", time, inputs)

        self._default_inputs()
        self._mosaik_inputs(inputs)

        for eid, model in self.models.items():
            model.step()

        self.now_dt += timedelta(seconds=self.step_size)
        self._generate_flexibilities()

        return time + self.step_size

    def get_data(self, outputs):
        """Returns the requested outputs (if feasible)"""

        data = dict()
        for eid, attrs in outputs.items():
            data[eid] = dict()
            model = eid.split("-")[0]
            for attr in attrs:

                if attr == "flexibilities":
                    value = self.models[eid].flexibilities
                elif attr == "schedule":
                    value = self.models[eid].schedule._data.loc[
                        self.now_dt : self.now_dt
                        + timedelta(hours=self.forecast_horizon_hours)
                        - timedelta(seconds=self.step_size)
                    ]
                elif attr == "target":
                    value = self.models[eid].schedule._data.loc[
                        self.now_dt - timedelta(seconds=self.step_size)
                    ]["target"]

                else:
                    # Apply correction of attr if necessary
                    if attr in ("p_mw", "p_th_mw", "q_mvar"):
                        attr = attr.replace("m", "k")

                    value = getattr(self.models[eid].state, attr)

                    # Apply correction of the value if necessary
                    if attr in ("p_kw", "p_th_kw", "q_kvar"):
                        attr = attr.replace("k", "m")
                        value *= 1e-3

                data.setdefault(eid, dict())[attr] = value

        LOG.debug("Gathered outputs: %s.", data)
        return data

    def _update_meta(self):
        for model in self.meta["models"].keys():
            self.meta["models"][model]["attrs"].append("flexibilities")
            self.meta["models"][model]["attrs"].append("schedule")
            self.meta["models"][model]["attrs"].append("target")

        self.meta["models"]["Photovoltaic"]["attrs"] += [
            "forecast_t_air_deg_celsius",
            "forecast_bh_w_per_m2",
            "forecast_dh_w_per_m2",
        ]
        self.meta["models"]["CHP"]["attrs"] += [
            "forecast_day_avg_t_air_deg_celsius"
        ]
        self.meta["models"]["HVAC"]["attrs"] += ["forecast_t_air_deg_celsius"]

    def _default_inputs(self):
        for eid, model in self.models.items():
            model.inputs.step_size = self.step_size
            model.inputs.now = self.now_dt

    def _mosaik_inputs(self, inputs):

        for eid, attrs in inputs.items():
            for attr, src_ids in attrs.items():

                if "forecast" in attr:
                    for forecast in src_ids.values():
                        self.models[eid].update_forecasts(forecast)
                    continue
                elif attr == "schedule":
                    for schedule in src_ids.values():
                        if schedule is not None:
                            self.models[eid].update_schedule(schedule)
                    continue

                # Aggregate inputs from different sources
                attr_sum = 0
                for val in src_ids.values():
                    if val is not None:
                        attr_sum += val
                attr_sum /= len(src_ids)

                if attr == "set_percent":
                    self.models[eid].set_percent = attr_sum
                else:
                    # Apply correction if necessary
                    if attr in ("p_mw", "p_th_mw", "q_mvar"):
                        attr = attr.replace("m", "k")
                        attr_sum *= 1e3

                    setattr(self.models[eid].inputs, attr, attr_sum)

    def _generate_flexibilities(self):
        if self.provide_flexibilities:
            for eid, model in self.models.items():
                model.generate_schedules(
                    self.now_dt.strftime(GER),
                    self.flexibility_horizon_hours,
                    self.num_schedules,
                )


if __name__ == "__main__":
    mosaik_api.start_simulation(FlexibilitySimulator())



"""This module contains the :class:`.Schedule` that is used by the
:class:`.ForecastModel` and :class:`.FlexibilityModel`.

"""
from datetime import datetime, timedelta
import numpy as np
import pandas as pd

from . import LOG


class Schedule:
    """Schedule class for the :class:`.ForecastModel`.

    Calling the schedule obj like this::

        df = schedule()

    returns the data frame that stores the schedule.


    Parameters
    ----------

    """

    def __init__(
        self,
        start_date,
        step_size,
        pname,
        qname,
        forecast_horizon_hours=1,
    ):
        self._data = None
        self._forecast_horizon_hours = forecast_horizon_hours
        self._pname = pname
        self._qname = qname

        self.now_dt = start_date
        self.step_size = step_size

    def init(self):
        """Initialize the schedule data frame.

        After initialization, the schedule's data frame should have
        three columns *target*, *p_kw*, and *q_kvar* and a number of
        rows, each value initialized with np.nan.

        The number of rows is defined by the
        :attr:`_forecast_horizon_hours` as seconds divided by the
        :attr:`step_size`.

        """
        LOG.debug("Creating new schedule dataframe")
        index = pd.date_range(
            self.now_dt,
            self.now_dt
            + timedelta(hours=self._forecast_horizon_hours)
            - timedelta(seconds=self.step_size),
            freq=f"{self.step_size}S",
        )

        self._data = pd.DataFrame(
            columns=["target", self._pname, self._qname], index=index
        )

    def update(self, other):
        """Update this schedules' data with another dataframe.

        Parameters
        ----------
        other : pandas.DataFrame
            A dataframe with datetime as index and values for the
            columns *"target"*, *"p_kw"*, and *"q_kvar"*. Note that
            :attr:`pname` and :attr:`qname` match in both data frames.

        """
        if self._data is None:
            self.init()

        for index, _ in other.iterrows():
            if index not in self._data.index:
                break

        self._data.update(other.loc[:index])
        self._data = self._data.append(other.loc[index:])
        self._data = self._data[~self._data.index.duplicated()]

    def update_row(self, index, setpoint, p_val, q_val):
        if index in self._data.index:
            self._data.loc[index]["target"] = setpoint
            self._data.loc[index][self._pname] = p_val
            self._data.loc[index][self._qname] = q_val
        else:
            self._data = self._data.append(
                pd.DataFrame(
                    data={
                        "target": setpoint,
                        self._pname: p_val,
                        self._qname: q_val,
                    },
                    index=[index],
                )
            )

    def update_entry(self, index, col, val):
        self._data.loc[index][col] = val

    def has_index(self, index):
        return index in self._data.index

    def reschedule_required(self):
        """Check if a reschedule is required.

        The current schedule is checked for the next hours, specified
        by :attr:`_forecast_horizon_hours*. A reschedule is required
        when on of the indices is missing of one of the values within
        the limit is np.nan.

        Returns
        -------
        bool
            *True* when a reschedule is required and *False* otherwise.
        """
        now = self.now_dt + timedelta(seconds=self.step_size)
        limit = int(self._forecast_horizon_hours * 3_600 / self.step_size)
        reschedule = False

        for counter in range(limit):
            if now not in self._data.index:
                return True
            elif np.isnan(self._data.loc[now]["target"]):
                return True
            elif np.isnan(self._data.loc[now][self._pname]):
                return True
            elif np.isnan(self._data.loc[now][self._qname]):
                return True
            else:
                now += timedelta(seconds=self.step_size)

        return False

    def prune(self):
        self._data = self._data.loc[self.now_dt :]

    def get(self, col, now=None):
        """Return the value for a certain column.

        If no time index is provided, the current datetime object of
        the schedule is used.

        Returns
        -------
        np.float
            The value of the specified column at index *now*.

        """
        if now is None:
            now = self.now_dt

        return self._data.loc[now][col]

    def __call__(self):
        return self._data

    def from_dataframe(self, dataframe):
        # TODO:
        pass



"""This module contains an inverter model"""

import math
from copy import copy

from pysimmods.other.invertersim.config import InverterConfig
from pysimmods.other.invertersim.inputs import InverterInputs
from pysimmods.other.invertersim.state import InverterState
from pysimmods.model.model import Model


class Inverter(Model):
    """Inverter model

    Supports five different modes:

    - *p_set*: p_set_kw is provided, cos phi remains constant and
      q_kvar is calculated with sqrt(s^2-p^2).
    - *q_set*: q_set_kvar is provided, cos phi remains constant and
      p_kw is calculated with sqrt(s^2-q^2)
    - *cos_phi_set*: cos phi is provided, p_kw is calculated from
      p_in_kw and q_kvar is calculated accordingly.
    - *pq_set*: p_set_kw and q_set_kvar are provided, with p being
      prioritized, i.e., q is limited by sqrt(s^2-p^2), and cos phi is
    calculated with p/s.
    - *qp_set*: q_set_kvar and p_set_kw are provided, with q being
      prioritized, i.e., p is limited by sqrt(s^2-q^2), and cos phi is
      calculated with p/s

    Parameters
    ----------
    params : dict
        A *dict* containing configuration parameters.
    inits : dict, optional
        A *dict* containing initialization parameters.

    Attributes
    ----------
    config : :class:`~.InverterConfig`
        Configuration parameters of the inverter.
    state : :class:`~.InverterState`
        Initialization parameters of the inverter.
    inputs : :class:`~.InverterInputs`
        Input parameters of the inverter.

    """

    def __init__(self, params, inits=None):
        if inits is None:
            inits = dict()
        self.config = InverterConfig(params)
        self.state = InverterState(inits)
        self.inputs = InverterInputs()

    def step(self):
        """Perform simulation step"""
        next_state = copy(self.state)

        self._check_inputs()

        if self.config.q_control == "p_set":
            self._calc_p_set(next_state)

        elif self.config.q_control == "q_set":
            self._calc_q_set(next_state)

        elif self.config.q_control == "cos_phi_set":
            self._calc_cos_phi_set(next_state)

        elif self.config.q_control == "pq_set":
            self._calc_pq_set(next_state)

        elif self.config.q_control == "qp_set":
            self._calc_qp_set(next_state)

        if self.config.inverter_mode == "inductive":
            next_state.q_kvar *= -1

        next_state.p_kw *= self.config.gsign
        next_state.q_kvar *= self.config.gsign

        self.state = next_state
        self.inputs.reset()

    def _check_inputs(self):
        if self.inputs.p_set_kw is None:
            self.inputs.p_set_kw = self.inputs.p_in_kw

        if self.inputs.q_set_kvar is None:
            self.inputs.q_set_kvar = self.config.sn_kva

        if self.inputs.cos_phi_set is None:
            self.inputs.cos_phi_set = self.config.cos_phi

    def _calc_p_set(self, next_state):

        # Current active power
        p_kw = min(abs(self.inputs.p_in_kw), abs(self.inputs.p_set_kw))

        # Check constraints for apparent power
        s_kva = p_kw / self.config.cos_phi
        s_kva = min(s_kva, self.config.sn_kva)

        # Check constraints for active power
        p_max_kw = s_kva * self.config.cos_phi
        p_kw = min(p_kw, p_max_kw)

        # Current reactive power
        q_kvar = (s_kva ** 2 - p_kw ** 2) ** 0.5
        q_max_kvar = (self.config.sn_kva ** 2 - p_kw ** 2) ** 0.5
        q_kvar = min(q_max_kvar, q_kvar)

        # Update state
        next_state.p_kw = p_kw
        next_state.q_kvar = q_kvar
        next_state.cos_phi = self.config.cos_phi

    def _calc_q_set(self, next_state):

        # Check constraints for reactive power
        phi = math.acos(self.config.cos_phi)
        q_max_kvar = self.config.sn_kva * math.sin(phi)
        q_kvar = min(abs(self.inputs.q_set_kvar), q_max_kvar)

        # Check constraints for active power
        p_max_kw = (self.config.sn_kva ** 2 - q_kvar ** 2) ** 0.5
        p_kw = min(abs(self.inputs.p_in_kw), p_max_kw)

        # Update state
        next_state.q_kvar = q_kvar
        next_state.p_kw = p_kw
        next_state.cos_phi = self.config.cos_phi

    def _calc_cos_phi_set(self, next_state):

        # Check cos phi input

        # Check constraints for apparent power
        s_kva = abs(self.inputs.p_in_kw) / self.inputs.cos_phi_set
        s_kva = min(s_kva, self.config.sn_kva)

        # Check constraints for active power
        p_max_kw = s_kva * self.inputs.cos_phi_set
        p_kw = min(abs(self.inputs.p_in_kw), p_max_kw)

        # Calculate reactive power
        q_kvar = (s_kva ** 2 - p_kw ** 2) ** 0.5

        # Update state
        next_state.p_kw = p_kw
        next_state.q_kvar = q_kvar
        next_state.cos_phi = self.inputs.cos_phi_set

    def _calc_pq_set(self, next_state):

        # Calculate active power
        p_kw = min(abs(self.inputs.p_in_kw), abs(self.inputs.p_set_kw))

        # Check constraints for reactive power
        q_max_kvar = (self.config.sn_kva ** 2 - p_kw ** 2) ** 0.5
        q_kvar = min(q_max_kvar, abs(self.inputs.q_set_kvar))

        # Calculate cos phi
        cos_phi = p_kw / self.config.sn_kva

        # Update state
        next_state.p_kw = p_kw
        next_state.q_kvar = q_kvar
        next_state.cos_phi = cos_phi

    def _calc_qp_set(self, next_state):

        # Calculate reactive power
        q_kvar = abs(self.inputs.q_set_kvar)
        q_kvar = min(q_kvar, self.config.sn_kva)

        # Calculate active power
        p_kw = min(abs(self.inputs.p_in_kw), abs(self.inputs.p_set_kw))

        # Check constraints for active power
        p_max_kw = (
            self.config.sn_kva ** 2 - self.inputs.q_set_kvar ** 2
        ) ** 0.5
        p_kw = min(p_max_kw, p_kw)

        # Calculate cos phi
        cos_phi = p_kw / self.config.sn_kva

        # Update state
        next_state.p_kw = p_kw
        next_state.q_kvar = q_kvar
        next_state.cos_phi = cos_phi



"""This module contains the base class for all pysimmods models."""
import copy
from abc import ABC, abstractmethod


class Model(ABC):
    """Base class for pysimmods models.

    Parameters
    ----------
    params : dict
        A *dict* containing configuration parameters.
    inits : dict
        A *dict* containing the state variables.

    Attributes
    ----------
    config : pysimmods.model.config:Config
        A config object holding all the configuration parameters of
        this model. These do not change during simulation.
    state : pysimmods.model.state:State
        A state object holding all variable parameters of this model.
        These values normally change during each step.
    inputs : pysimmods.model.inputs:Inputs
        An inputs object defining the inputs for this model. In each
        step, all the inputs need to be provided.

    """

    def __init__(self, params, inits):
        self.config = None
        self.state = None
        self.inputs = None

    @abstractmethod
    def step(self):
        """Perform a simulation step."""
        raise NotImplementedError

    def get_state(self):
        """Return the current state of the model

        Returns
        -------
        dict
            The current state of the model in form of a dictionary
            containing entries for all state variables. Returned dict
            can be assigned to the *inits* argument when creating a new
            model instance.

        """

        try:
            return {
                attr: getattr(self.state, attr)
                for attr in self.state.__slots__
            }
        except AttributeError:
            return copy.deepcopy(self.state.__dict__)

    def set_state(self, state):
        """Set the current state of the model.

        Parameters
        ----------
        state : dict
            A *dict* containing entries for all state variables.

        """
        for attr, value in state.items():
            setattr(self.state, attr, value)

    @property
    def p_max_kw(self):
        return self.config.p_max_kw

    @property
    def p_min_kw(self):
        return self.config.p_min_kw

    @property
    def p_kw(self):
        return self.state.p_kw

    @property
    def q_max_kvar(self):
        return self.config.q_max_kvar

    @property
    def q_min_kvar(self):
        return self.config.q_min_kvar

    @property
    def q_kvar(self):
        return self.state.q_kvar

    @property
    def set_percent(self):
        raise NotImplementedError

    @set_percent.setter
    def set_percent(self, set_percent):
        raise NotImplementedError

    @property
    def default_schedule(self):
        """Return the default schedule.

        Returns
        -------
        list
            A *list* containing a default schedule with percentage
            power for each hour of the day (i.e.,
            len(default_schedule) == 24). This is used if no other *p*
            or *q* input is provided.

        """
        return self.config.default_schedule



"""This module contains a :class:`mosaik_api.Simulator` for all models
of the pysimmods package.

"""
from datetime import datetime, timedelta
import mosaik_api

from pysimmods.mosaik import LOG
from pysimmods.util.dateformat import GER

from .meta import META, MODELS


class PysimmodsSimulator(mosaik_api.Simulator):
    """The Pysimmods simulator. """

    def __init__(self):
        super().__init__(META)
        self.sid = None
        self.models = dict()
        self.num_models = dict()

        self.step_size = None
        self.now_dt = None

    def init(self, sid, **sim_params):
        """Called exactly ones after the simulator has been started.

        Parameters
        ----------
        sid : str
            Simulator ID for this simulator.
        start_date : str
            The start date as UTC ISO 8601 date string.
        step_size : int, optional
            Step size for this simulator. Defaults to 900.

        Returns
        -------
        dict
            The meta dict (set by *mosaik_api.Simulator*).

        """
        self.sid = sid
        if "step_size" not in sim_params:
            LOG.debug(
                "Param *step_size* not provided. "
                "Using default step size of 900."
            )
        self.step_size = sim_params.get("step_size", 900)
        self.now_dt = datetime.strptime(sim_params["start_date"], GER)

        return self.meta

    def create(self, num, model, **model_params):
        """Initialize the simulation model instance (entity).

        Parameters
        ----------
        num : int
            The number of models to create.
        model : str
            The name of the models to create. Must be present inside
            the simulator's meta.

        Returns
        -------
        list
            A list with information on the created entity.

        """
        entities = list()
        params = model_params["params"]
        inits = model_params["inits"]
        self.num_models.setdefault(model, 0)

        for _ in range(num):

            eid = f"{model}-{self.num_models[model]}"
            self.models[eid] = MODELS[model](params, inits)
            self.num_models[model] += 1
            entities.append({"eid": eid, "type": model})

        return entities

    def step(self, time, inputs):
        """Perform a simulation step.

        Parameters
        ----------
        time : int
            The current simulation step (by convention in seconds since
            simulation start.
        inputs : dict
            A *dict* containing inputs for entities of this simulator.

        Returns
        -------
        int
            The next step this simulator wants to be stepped.

        """
        LOG.debug("At step %d: Received inputs: %s.", time, inputs)
        # Set default inputs
        for eid, model in self.models.items():
            model.inputs.step_size = self.step_size
            model.inputs.now = self.now_dt

        # Set inputs from other simulators
        for eid, attrs in inputs.items():
            for attr, src_ids in attrs.items():

                # Aggregate inputs from different sources
                attr_sum = 0
                for src_id, val in src_ids.items():
                    if val is not None:
                        attr_sum += val
                attr_sum /= len(src_ids)

                # Set the inputs
                if attr == "set_percent":
                    self.models[eid].set_percent = attr_sum
                else:
                    # Apply correction if necessary
                    if attr in ("p_set_mw", "p_th_set_mw", "q_set_mvar"):
                        attr = attr.replace("m", "k")
                        attr_sum *= 1e3

                    setattr(self.models[eid].inputs, attr, attr_sum)

        # Step the models
        for model in self.models.values():
            model.step()

        # Update time for the next step
        self.now_dt += timedelta(seconds=self.step_size)

        return time + self.step_size

    def get_data(self, outputs):
        """Return the requested output (if feasible).

        Parameters
        ----------
        outputs : dict
            A *dict* containing requested outputs of each entity.

        Returns
        -------
        dict
            A *dict* containing the values of the requested outputs.

        """

        data = dict()
        for eid, attrs in outputs.items():
            for attr in attrs:
                # Apply correction of the attr if necessary
                if attr in ("p_mw", "p_th_mw", "q_mvar"):
                    attr = attr.replace("m", "k")

                value = getattr(self.models[eid].state, attr)

                # Apply correction of the value if necessary
                if attr in ("p_kw", "p_th_kw", "q_kvar"):
                    attr = attr.replace("k", "m")
                    value *= 1e-3

                data.setdefault(eid, dict())[attr] = value

        LOG.debug("Gathered outputs: %s.", data)
        return data


if __name__ == "__main__":
    mosaik_api.start_simulation(PysimmodsSimulator())



"""This module contains the :class:`ForecastModel`."""
from datetime import datetime, timedelta

import numpy as np

import pandas as pd
from pysimmods.util.dateformat import GER
from pysimmods.other.flexibility.schedule import Schedule
from . import LOG


class ForecastModel:
    """The forecast model for all pysimmods.

    Parameters
    ----------
    model : :class:`.Model`
        An instance of a subclass of the base model class (e.g.,
        :class:`.Battery` model).
    start_date : str or datetime
        An ISO datetime string or a datetime object defining the start
        date.
    step_size : int, optional
        The step size of the model in seconds. Will be used as fallback
        if no step size is provided to the model. Defaults to 900.

    """

    def __init__(
        self,
        model,
        start_date,
        step_size=900,
        forecast_horizon_hours=1,
        flexibility_horizon_hours=3,
        num_schedules=10,
        seed=None,
        unit="kw",
        percent_factor=1.0,
        priorize_schedule=False,
    ):
        self.model = model
        """A reference to the model"""

        if isinstance(start_date, str):
            self.now_dt = datetime.strptime(start_date, GER)
        else:
            self.now_dt = start_date
        """The current local time."""

        self.step_size = step_size
        """The step size of the model."""

        self.percent_factor = percent_factor
        self.unit_factor = 1.0
        self.pname = "p_kw"
        self.qname = "q_kvar"
        self.priorize_schedule = priorize_schedule
        self.forecasts = None
        """A dictionary containing forecasts for the inputs of 
        the underlying model.
        """
        self.flexibilities = None
        """A dictionary containing the current flexibilities of
        the underlying model.
        """
        self.forecast_horizon_hours = forecast_horizon_hours
        self.num_schedules = num_schedules

        if unit == "mw":
            self.unit_factor = 0.001
            self.pname = "p_mw"
            self.qname = "q_mvar"
        elif unit == "w":
            self.unit_factor = 1000.0
            self.pname = "p_w"
            self.qname = "q_var"

        self.schedule = Schedule(
            self.now_dt,
            self.step_size,
            self.pname,
            self.qname,
            self.forecast_horizon_hours,
        )
        """A dictionary containing the schedule for the model."""
        self.active_schedule_id = None
        """A pointer to the currently active schedule."""

        if seed is not None:
            self.rng = np.random.RandomState(seed)
        else:
            self.rng = np.random.RandomState()

    def update_forecasts(self, forecasts):
        if self.forecasts is None:
            self.forecasts = forecasts
        else:
            for col in forecasts.columns:
                if col not in self.forecasts.columns:
                    self.forecasts[col] = np.nan
            for index, _ in forecasts.iterrows():
                if index not in self.forecasts.index:
                    break

            self.forecasts.update(forecasts.loc[:index])
            self.forecasts = self.forecasts.append(forecasts.loc[index:])

            self.forecasts = self.forecasts[~self.forecasts.index.duplicated()]

    def update_schedule(self, schedule):
        self.schedule.update(schedule)

    def _get_setpoint(self):
        try:
            schedule_set = self.schedule.get("target") / self.percent_factor
        except TypeError:
            schedule_set = None
        try:
            model_set = self.model.set_percent * self.percent_factor
        except TypeError:
            model_set = None

        default_set = (
            self.model.default_schedule[self.now_dt.hour] * self.percent_factor
        )

        priority = [model_set, schedule_set]

        if self.priorize_schedule:
            priority = priority[::-1]

        priority.append(default_set)

        setpoint = None
        for setval in priority:
            if setval is not None and ~np.isnan(setval):
                setpoint = setval
                break
        else:
            raise ValueError("Setpoint for model %s not set.", self.model)

        return setpoint

    def step(self):
        """Perform a simulation step of the underlying model.

        Also updates the internal state of the flexibility model.

        """
        # Update the step size
        self._check_inputs()

        setpoint = self._get_setpoint()
        self.set_percent = setpoint
        self.model.step()

        self.schedule.update_row(
            self.now_dt,
            setpoint,
            self.model.p_kw * self.unit_factor,
            self.model.q_kvar * self.unit_factor,
        )
        self._check_schedule()

        self.now_dt += timedelta(seconds=self.step_size)
        self.schedule.now_dt = self.now_dt

    def _check_inputs(self):
        if self.model.inputs.step_size is not None:
            self.step_size = self.model.inputs.step_size
        else:
            self.model.inputs.step_size = self.step_size
        self.model.inputs.now = self.now_dt

        if self.schedule() is None:
            self.schedule.init()

    def _check_schedule(self):

        if self.schedule.reschedule_required():
            self._create_init_schedule()
            self.schedule.prune()

    def _create_init_schedule(self):
        state_backup = self.model.get_state()

        now = self.now_dt + timedelta(seconds=self.step_size)
        periods = int(self.forecast_horizon_hours * 2 * 3_600 / self.step_size)

        for _ in range(periods):
            if not self.schedule.has_index(now):
                self.schedule.update_row(now, np.nan, np.nan, np.nan)
            default_setpoint = (
                self.model.default_schedule[now.hour] * self.percent_factor
            )
            if np.isnan(self.schedule.get("target", now)):
                self.schedule.update_entry(now, "target", default_setpoint)

            try:
                self._perform_step(now, self.schedule.get("target", now))

                self.schedule.update_entry(
                    now, self.pname, self.model.p_kw * self.unit_factor
                )
                self.schedule.update_entry(
                    now, self.qname, self.model.q_kvar * self.unit_factor
                )

            except KeyError:
                # Forecast is missing
                LOG.warning("No forecast provided for model %s.", self.model)
                self.schedule.update_row(now, np.nan, np.nan, np.nan)

            now += timedelta(seconds=self.step_size)

        self.model.set_state(state_backup)

    def _perform_step(self, index, set_percent):
        self.set_percent = set_percent

        if self.forecasts is not None:
            for col in self.forecasts.columns:
                if hasattr(self.model.inputs, col):
                    setattr(
                        self.model.inputs, col, self.forecasts.loc[index, col]
                    )

        self.model.inputs.now = index
        self.model.inputs.step_size = self.step_size
        self.model.step()

    @property
    def inputs(self):
        return self.model.inputs

    @property
    def config(self):
        return self.model.config

    @property
    def state(self):
        return self.model.state

    @property
    def set_percent(self):
        return self.model.set_percent * self.percent_factor

    @set_percent.setter
    def set_percent(self, set_percent):
        self.model.set_percent = set_percent / self.percent_factor



"""This module contains the base class for all model inputs."""
from datetime import datetime
from pysimmods.util.dateformat import GER


class ModelInputs:
    """Base class for model inputs.

    Attributes
    ----------
    p_set_kw : float
        Target electrical activate power in [kW].
    q_set_kvar : float
        Target electrical reactive power in [kVAr].
    step_size : int
        Step size for the net step in [s].

    """

    def __init__(self):
        self.p_set_kw = None
        self.q_set_kvar = None
        self.step_size = None
        self._now = None

    @property
    def now(self):
        return self._now

    @now.setter
    def now(self, value):
        """Set the current time.

        Parameters
        ----------
        value : datetime.datetime
            The time to set. If *value* is an UTC ISO 8601 time string
            or an unix timestamp, it will be converted to datetime.

        Raises
        ------
        ValueError
            If *value* is neither datetime.datetime nor an UTC ISO 8601
            time string nor an unix timestamp.

        """

        if isinstance(value, str):
            self._now = datetime.strptime(value, GER)
        elif isinstance(value, int):
            self._now = datetime.utcfromtimestamp(value)
        elif isinstance(value, datetime):
            self._now = value
        else:
            raise ValueError(f"Unknown date format: {type(value)}")

    def reset(self):
        """To be called at the end of each step."""
        for attr in self.__dict__.keys():
            setattr(self, attr, None)



"""This module contains the *mosaik_api* meta definition for pysimmods
models.
"""
from pysimmods.buffer.batterysim import Battery
from pysimmods.consumer.hvacsim import HVAC
from pysimmods.generator.chplpgsystemsim import CHPLPGSystem
from pysimmods.generator.pvsystemsim import PVPlantSystem

META = {
    "models": {
        "Battery": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": [
                "set_percent",
                "p_mw",
                "q_mvar",
                "soc_percent",
            ],
        },
        "Photovoltaic": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": [
                "set_percent",
                "p_mw",
                "q_mvar",
                "t_air_deg_celsius",
                "t_module_deg_celsius",
                "bh_w_per_m2",
                "dh_w_per_m2",
                "s_module_w_per_m2",
            ],
        },
        "CHP": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": [
                "set_percent",
                "p_mw",
                "q_mvar",
                "day_avg_t_air_deg_celsius",
                "p_th_mw",
            ],
        },
        "HVAC": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": [
                "set_percent",
                "p_mw",
                "q_mvar",
                "t_air_deg_celsius",
            ],
        },
        "Biogas": {
            "public": True,
            "params": ["params", "inits"],
            "attrs": ["set_percent", "p_mw", "q_mvar", "p_th_mw"],
        },
    }
}

MODELS = {
    "Photovoltaic": PVPlantSystem,
    "HVAC": HVAC,
    "Battery": Battery,
    "CHP": CHPLPGSystem,
}



"""This module contains a model of a pv system with pv modules and an
inverter"""
from pysimmods.other.invertersim.config import InverterConfig
from pysimmods.other.invertersim.inverter import Inverter
from pysimmods.model.generator import Generator
from pysimmods.generator.pvsim.pvp import PhotovoltaicPowerPlant
from pysimmods.generator.pvsystemsim.config import PVSystemConfig
from pysimmods.generator.pvsystemsim.inputs import PVSystemInputs
from pysimmods.generator.pvsystemsim.state import PVSystemState


class PVPlantSystem(Generator):
    """Pv system with pv modules and inverter"""

    def __init__(self, params, init_vals):
        self.config = PVSystemConfig(params)
        self.inputs = PVSystemInputs()
        self.state = PVSystemState(init_vals)
        self.pv = PhotovoltaicPowerPlant(params["pv"], init_vals["pv"])
        self.inverter = Inverter(params["inverter"])

    def step(self):
        """Perform simulation step"""

        # Step the pv plant
        self.pv.inputs.bh_w_per_m2 = self.inputs.bh_w_per_m2
        self.pv.inputs.dh_w_per_m2 = self.inputs.dh_w_per_m2
        self.pv.inputs.s_module_w_per_m2 = self.inputs.s_module_w_per_m2
        self.pv.inputs.t_air_deg_celsius = self.inputs.t_air_deg_celsius
        self.pv.inputs.step_size = self.inputs.step_size
        self.pv.inputs.now = self.inputs.now

        self.pv.step()

        # Step the inverter
        self.inverter.inputs.p_in_kw = self.pv.state.p_kw
        self.inverter.inputs.p_set_kw = self.inputs.p_set_kw
        self.inverter.inputs.q_set_kvar = self.inputs.q_set_kvar
        self.inverter.inputs.cos_phi_set = self.inputs.cos_phi_set

        self.inverter.step()

        # Update state
        self.state.t_module_deg_celsius = self.pv.state.t_module_deg_celsius
        self.state.p_kw = self.inverter.state.p_kw
        self.state.q_kvar = self.inverter.state.q_kvar
        self.state.cos_phi = self.inverter.state.cos_phi

        self.inputs.reset()

    def get_state(self):
        """Get state"""
        state_dict = {
            "pv": self.pv.get_state(),
            "inverter": self.inverter.get_state(),
        }
        return state_dict

    def set_state(self, state_dict):
        """Set state"""
        self.pv.set_state(state_dict["pv"])
        self.inverter.set_state(state_dict["inverter"])

        self.state.t_module_deg_celsius = self.pv.state.t_module_deg_celsius
        self.state.p_kw = self.inverter.state.p_kw
        self.state.q_kvar = self.inverter.state.q_kvar

    @property
    def set_percent(self):
        p_dominated = ["p_set", "pq_set"]
        q_dominated = ["q_set", "qp_set"]
        cos_phi_dominated = ["cos_phi_set"]
        mode = self.config.inverter.q_control

        if mode in p_dominated:
            p_kw = self.inputs.p_set_kw
            if p_kw is None:
                p_kw = self.state.p_kw

            return 100 * abs(
                p_kw / abs(self.config.p_max_kw - self.config.p_min_kw)
            )

        if mode in q_dominated:
            q_kvar = self.inputs.q_set_kvar
            if q_kvar is None:
                q_kvar = self.state.q_kvar

            return 100 * abs(
                q_kvar / abs(self.config.q_max_kvar - self.config.q_min_kvar)
            )

        if mode in cos_phi_dominated:
            cos_phi = self.inputs.cos_phi_set
            if cos_phi is None:
                cos_phi = self.state.cos_phi
            return cos_phi * 100

    @set_percent.setter
    def set_percent(self, set_percent):
        mode = self.config.inverter.q_control

        set_percent = max(min(abs(set_percent), 100.0), 0.0)
        if mode == "p_set":
            bounds = [abs(self.config.p_max_kw), abs(self.config.p_min_kw)]
            self.inputs.p_set_kw = min(bounds) + set_percent * 0.01 * (
                max(bounds) - min(bounds)
            )

        elif mode == "q_set":
            bounds = [abs(self.config.q_max_kvar), abs(self.config.q_min_kvar)]
            self.inputs.q_set_kvar = min(bounds) + set_percent * 0.01 * (
                max(bounds) - min(bounds)
            )

        elif mode == "pq_set":
            bounds = [abs(self.config.p_max_kw), abs(self.config.p_min_kw)]
            self.inputs.p_set_kw = min(bounds) + set_percent * 0.01 * (
                max(bounds) - min(bounds)
            )
            self.inputs.q_set_kvar = (
                self.config.inverter.sn_kva ** 2 - self.inputs.p_set_kw ** 2
            ) ** 0.5

        elif mode == "qp_set":
            bounds = [abs(self.config.q_max_kvar), abs(self.config.q_min_kvar)]
            self.inputs.q_set_kvar = min(bounds) + set_percent * 0.01 * (
                max(bounds) - min(bounds)
            )
            self.inputs.p_set_kw = (
                self.config.inverter.sn_kva ** 2 - self.inputs.q_set_kvar ** 2
            ) ** 0.5

        elif mode == "cos_phi_set":
            self.inputs.cos_phi_set = set_percent * 0.01



"""This module contains the base class for all model states."""


class ModelState:
    """Base class for model states.

    Parameters
    ----------
    inits : dict
        A *dict* containing the state variables.

    Attributes
    ----------
    p_kw : float
        Current (or last) electrical active power in [kW].
    q_kvar : float
        Current (or last) electrical reactive power in [kVAr].
    """

    def __init__(self, inits):
        self.p_kw = inits.get("p_kw", 0)
        self.q_kvar = inits.get("q_kvar", 0)
